/*
 * in.timed	This file contains an implementation of the "server"
 *		for the TIMEP protocol, as described in RFC 868.  This
 *		protocol can be used to ask servers like this one for
 *		their current time-and-day.
 *
 * Usage:	in.timed [-v]
 *
 * Version:	@(#)in.timed.c		2.00	08/11/98
 *
 * Authors:	Michael Temari, <temari@temari.ae.ge.com>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *		BALATON Zoltan, <balaton@eik.bme.hu>
 * 
 * Changelog:
 * 
 * Sun Nov  8 21:53:25 CET 1998  BALATON Zoltan  <balaton@eik.bme.hu>
 * 
 * 	* Cleaned up and rewritten the original in.timed.c 1.10 (12/12/93)
 * 	* Added UDP support in compliance with the RFC.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

/* 2208988800 is 00:00 1 Jan 1970 GMT (the unix time Epoch) in nettime */
#define EPOCH  2208988800UL
/* Check error condition and return on it */
#define CHKERR(x)  if( (x) < 0 ) return (-1)

char *Version = "@(#)timed 2.00 (08/11/98)";
int opt_v = 0;	/* debugging output flag */

/* Set the O_NONBLOCK flag of FD if SET is nonzero, or clear it otherwise. */
/* Return 0 on success, or -1 on error with errno set. */
static int set_nonblock_flag(int fd, int set)
{
  int flags;

  CHKERR(flags = fcntl(fd, F_GETFL, 0));
  flags = ( set ? flags | O_NONBLOCK : flags & ~O_NONBLOCK );
  return(fcntl(fd, F_SETFL, flags));
}

/* Answer an incoming TIMEP request on inet socket S. */
/* Return 0 on success, or -1 on error with errno set. */
static int timed(int s)
{
  struct sockaddr_in addr;
  int len, type;
  unsigned long val;

  len = sizeof(type);
  CHKERR(getsockopt(s, SOL_SOCKET, SO_TYPE, &type, &len));

  len = sizeof(addr);
  switch(type) {
    case SOCK_STREAM:
      CHKERR(getpeername(s, &addr, &len));
      break;
    case SOCK_DGRAM:
      CHKERR(set_nonblock_flag(s, 1));
      if( recvfrom(s, &val, sizeof(val), 0, &addr, &len) < 0 )
        return( errno == EWOULDBLOCK ? 0 : -1 );
      CHKERR(set_nonblock_flag(s, 0));
      break;
    default:
      errno = ESOCKTNOSUPPORT;
      return(-1);
  }

  val = htonl((unsigned long)time(NULL) + EPOCH);
  CHKERR(sendto(s, &val, sizeof(val), 0, &addr, len));

  return(0);
}

static void usage(void)
{
  fprintf(stderr, "Usage: in.timed [-v]\n");
  exit(-1);
}

int main(int argc, char *argv[])
{
  int c;

  opterr = 0;
  while((c = getopt(argc, argv, "v")) != EOF)
    switch(c) {
      case 'v':
        opt_v = 1;
        break;
      default:
        usage();
    }
  /* No more arguments allowed. */
  if(optind != argc) usage();

  /* Start the server. */
  c = timed(0);

  if( c && opt_v ) perror("in.timed");
  return(c);
}
